package checkers.igj;

import java.lang.annotation.Annotation;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

import javax.lang.model.element.*;
import javax.lang.model.type.*;

import checkers.igj.quals.*;
import checkers.types.*;

/**
 * This class encapsulate the relationships of immutability.
 */
public enum IGJImmutability {

    // ***************************************************
    // The Enum portion
    // ***************************************************
    READONLY(ReadOnly.class), IMMUTABLE(Immutable.class), MUTABLE(
            Mutable.class), ASSIGNSFIELDS(AssignsFields.class), PLACE_HOLDER(
            IGJPlaceHolder.class), WILD_CARD(W.class);

    private Class<? extends Annotation> annotation;

    private IGJImmutability(Class<? extends Annotation> annotation) {
        this.annotation = annotation;
    }

    /**
     * @return The {@code Annotation} that corresponds to this
     *         immutability.
     */
    public Class<? extends Annotation> getAnnotation() {
        return annotation;
    }

    /**
     * Finds the {@code IGJImmutability} type that corresponds to the
     * passed annotation and returns it, if one exists.
     * Otherwise, it throws an {@code IllegalArgumentException}.
     * 
     * @param annotation
     *            the name of the immutability {@code Annotation}
     * @return the {@code IGJImmutability} that corresponds to
     *         {@code annotation}
     */
    public static IGJImmutability valueOf(CharSequence annon) {
        String annotation = annon.toString();
        for (IGJImmutability value : values()) {
            if (value.getAnnotation().getCanonicalName().equals(annotation))
                return value;
        }

        throw new IllegalArgumentException(
                "Annotation is not an IGJ Annotation");
    }

    public static IGJImmutability valueOf(AnnotationData data) {
        Name qualifiedName = getAnnotationTypeName(data);
        return valueOf(qualifiedName);
    }

    /**
     * Compares this immutability type to otherType, and returns
     * {@code true} if this is a subtype of otherType.
     * 
     * @param otherType
     *            the immutability type to be compared to
     * @return true iff this <= othertype
     */
    public boolean isSubtypeOf(IGJImmutability otherType) {
        if (otherType == READONLY || otherType == PLACE_HOLDER
                || otherType == WILD_CARD)
            return true;
        else if (this == MUTABLE && otherType == ASSIGNSFIELDS)
            return true;
        else if (this == PLACE_HOLDER)
            return true;
        else
            return (this == otherType);
    }

    /**
     * Returns the immutability type of the type in the given
     * {@code AnnotationLocation}.
     * 
     * @param type
     *            the class type to be checked
     * @param location
     *            the location to be checked for immutability
     * @return the immutability type in the given location or
     *         {@code null} if immutability is not specified
     */
    public static IGJImmutability getIGJImmutabilityAt(
            AnnotatedClassType type, AnnotationLocation location) {
        AnnotationData annotation =
            getIGJAnnotationDataAt(type, location);
        if (annotation == null)
            return null;
        return valueOf(annotation);
    }

    // **************************************************************
    // Immutability Annotations Method
    // **************************************************************

    /**
     * Returns the specified IGJAnnotation at the specified location.
     * 
     * @param type
     * @param location
     * @return the specified IGJ Annotation found in the given
     *         location.
     */
    public static Class<? extends Annotation> getIGJAnnotationAt(
            AnnotatedClassType type, AnnotationLocation location) {

        if (type.hasAnnotationAt(ReadOnly.class, location))
            return ReadOnly.class;
        if (type.hasAnnotationAt(Mutable.class, location))
            return Mutable.class;
        if (type.hasAnnotationAt(Immutable.class, location))
            return Immutable.class;
        if (type.hasAnnotationAt(AssignsFields.class, location))
            return AssignsFields.class;
        if (type.hasAnnotationAt(W.class, location))
            return W.class;
        if (type.hasAnnotationAt(IGJPlaceHolder.class, location))
            return IGJPlaceHolder.class;

        // Doesn't have any annotation
        return null;
    }

    /**
     * Returns the specified IGJ AnnotationData at the specified
     * location
     * 
     * @param type
     * @param location
     * @return the specified IGJ AnnotationData found in the given
     *         location
     */
    public static AnnotationData getIGJAnnotationDataAt(
            AnnotatedClassType type, AnnotationLocation location) {
        Class<? extends Annotation> annotation =
            getIGJAnnotationAt(type, location);
        if (annotation == null)
            return null;
        Set<AnnotationData> annotations =
            type.getAnnotationData(annotation, true);
        assert annotations.size() < 2;
        if (annotations.isEmpty())
            return null;
        else
            return annotations.iterator().next();
    }

    public static Set<AnnotationData> getIGJAnnotationData(
            AnnotatedClassType type) {
        Set<AnnotationData> annotations = new HashSet<AnnotationData>();
        for (IGJImmutability annon : values()) {
            annotations.addAll(type.getAnnotationData(annon.getAnnotation()
                    .getCanonicalName(), true));
        }
        return annotations;
    }

    /**
     * Checks if type is mutable at the given location
     * 
     * @param type
     * @param location
     * @return true iff type is mutable at the given location
     */
    public static boolean isMutableAt(AnnotatedClassType type,
            AnnotationLocation location) {
        return (type.hasAnnotationAt(Mutable.class, location) || type
                .hasAnnotationAt(IGJPlaceHolder.class, location));
    }

    /**
     * Checks if the type is immutable at the given location
     * 
     * @param type
     * @param location
     * @return true iff type is immutable at the given location
     */
    public static boolean isImmutableAt(AnnotatedClassType type,
            AnnotationLocation location) {
        return (type.hasAnnotationAt(Immutable.class, location) || type
                .hasAnnotationAt(IGJPlaceHolder.class, location));
    }

    public static boolean isIGJAnnotation(CharSequence qualifiedName) {
        IGJImmutability[] values = IGJImmutability.values();
        for (int i = 0; i < values.length; ++i) {
            if (values[i].getAnnotation().getCanonicalName().equals(
                    qualifiedName))
                ;
            return true;
        }
        return false;
    }

    public static boolean isIGJAnnotation(AnnotationData data) {
        // TODO: Ease this code
        DeclaredType annonType = (DeclaredType) data.getType();
        TypeElement annonElem = (TypeElement) annonType.asElement();
        Name annotation = annonElem.getQualifiedName();

        return isIGJAnnotation(annotation);
    }

    /**
     * Utility method to extract the value of the wildcard Annotation
     * 
     * @param annotationData
     *            the IGJ annotation data
     * @return the id number of the annotationData
     */
    public static int getID(AnnotationData data) {
        // Defensive
        if (W.class.equals(data.getType().getClass()))
            throw new IllegalArgumentException("data is not a @W Annotation");

        // W has only one value
        // assert data.getValues().values().size() == 1; // assertion
        // is false due to a bug
        Map<? extends ExecutableElement, ? extends AnnotationValue> map =
            data.getValues();
        // Solve a bug
        if (map.isEmpty()) {
            return 0; // To be fixed
        } else
            return (Integer) data.getValues().values().iterator().next()
                    .getValue();
    }

    public static boolean isWildcard(AnnotationData data) {
        return (getAnnotationTypeName(data).contentEquals(W.class.getCanonicalName()));
    }

    public static Name getAnnotationTypeName(AnnotationData data) {
        DeclaredType type = (DeclaredType) data.getType();
        TypeElement element = (TypeElement) type.asElement();
        return element.getQualifiedName();
    }

    public static boolean isSubtype(AnnotationData child,
            AnnotationData parent) {
        assert isIGJAnnotation(child) && isIGJAnnotation(parent);
        if (getAnnotationTypeName(child).contentEquals(
                IGJPlaceHolder.class.getCanonicalName()))
            return true;
        else if (isWildcard(parent)) {
            return (isWildcard(child) && getID(child) == getID(parent));
        } else if (isWildcard(child)) {
            return READONLY.isSubtypeOf(valueOf(parent));
        }

        IGJImmutability childImmutabtility = valueOf(child);
        IGJImmutability parentImmutabtility = valueOf(parent);
        return childImmutabtility.isSubtypeOf(parentImmutabtility);
    }
}
