package checkers.types;

import java.lang.annotation.Annotation;
import java.util.Collections;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Set;

import javax.annotation.processing.ProcessingEnvironment;
import javax.lang.model.element.Element;
import javax.lang.model.element.TypeElement;
import javax.lang.model.type.*;

/**
 * Represents a annotated type in the JSR308 Java programming
 * language. Types include primitive types, declared types (class and
 * interface types), array types, type variables, and the null type.
 * Also represented are wildcard type arguments, the signature and
 * return types of executables, and pseudo-types corresponding to
 * packages and to the keyword void.
 * 
 * Types should be compared using the utility methods in
 * {@link AnnotatedTypes}. There is no guarantee that any particular
 * type will always be represented by the same object.
 * 
 * To implement operations based on the class of an TypeMirror object,
 * either use a visitor or use the result of the {@link getKind()}
 * method.
 * 
 */
public class AnnotatedTypeMirror {
    /** Processing Environment of the current round **/
    protected final ProcessingEnvironment env;

    protected final AnnotationFactory annotationFactory;
    
    /** Actual type wrapped with this AnnotatedTypeMirror **/
    protected final TypeMirror actualType;
    
    /** the Element associated with this instance value, if one exists **/
    // TODO: Clarify, with value not the element of the type.
    // I.e. For 'Integer i;' the element would be for 'i' not 'Integer'
    protected Element element;
    
    /** The enclosing Type **/
    protected AnnotatedTypeMirror enclosingType;
    
    protected final Set<AnnotationData> annotations =
        new HashSet<AnnotationData>();
    protected final Set<AnnotationData> includes =
        new HashSet<AnnotationData>();
    protected final Set<AnnotationData> excludes =
        new HashSet<AnnotationData>();

    /**
     * Constructor for AnnotatedTypeMirror.
     * 
     * @param type  the underlying type
     * @param env   Processing Environment
     */
    // TODO: Have static factory methods
    protected AnnotatedTypeMirror(TypeMirror type, ProcessingEnvironment env) {
        this.actualType = type;
        this.env = env;
        this.annotationFactory = new AnnotationFactory(env);
    }

    /**
     * Applies a visitor to this type.
     * 
     * @param <R>   the return type of the visitor's methods
     * @param <P>   the type of the additional parameter to the visitor's methods
     * @param v the visitor operating on this type
     * @param p additional parameter to the visitor 
     * @return  a visitor-specified result
     */
    public <R, P> R accept(AnnotatedTypeVisitor<R, P> v, P p) {
        return v.visitType(this, p);
    }

    /**
     * Returns the {@code kind} of this type
     * @return the kind of this type
     */
    public TypeKind getKind() {
        return actualType.getKind();
    }

    /**
     * Returns the underlying unannotated Java type wrapped with this
     * 
     * @return  the underlying type
     */
    public TypeMirror getUnderlyingType() {
        return actualType;
    }

    /**
     * Sets the enclosing type
     * 
     * @param enclosingType
     */
    public void setEnclosingType(AnnotatedTypeMirror enclosingType) {
        this.enclosingType = enclosingType;
    }
    
    /**
     * Returns the enclosing type, as in the type of {@code A} in the type
     * {@code A.B}.
     * 
     * @return enclosingType the enclosing type
     */
    public AnnotatedTypeMirror getEnclosingType() {
        return enclosingType;
    }

    /**
     * Returns the annotations on this type. 
     * 
     * It does not include annotations in deep types (type arguments, array 
     * components, etc).
     * 
     * @param implicit  whether included annotations are to be considered
     * @return  a set of the annotations on this
     */
    public Set<AnnotationData> getAnnotations(boolean implicit) {
        Set<AnnotationData> annotations = new HashSet<AnnotationData>();
        annotations.addAll(annotations);
        if (implicit) {
            annotations.addAll(includes);
            annotations.removeAll(excludes);
        }
        return Collections.unmodifiableSet(annotations);
    }

    /**
     * Determines whether the annotated type has an annotation with a certain name
     * 
     * @param annotationName the name of the annotation to check for
     * @param implicit  whether or not to check included/excluded annotations
     * @return true iff the annotation with the given name is pesent
     */
    public boolean hasAnnotation(CharSequence annotationName, boolean implicit) {
        AnnotationData annotation = 
            annotationFactory.createAnnotation(annotationName, 
                    AnnotationLocation.RAW);
        
        if (implicit) {
            if (excludes.contains(annotation))
                return false;
            
            if (includes.contains(annotation))
                return true;
        }
        
        return annotations.contains(annotation);
    }

    /**
     * Determines whether the annotated type has an annotation with a certain class
     * 
     * @param Annotation the class of the annotation to check for
     * @param implicit  whether or not to check included/excluded annotations
     * @return true iff the annotation with the given name is pesent
     */
    public boolean hasAnnotation(Class<? extends Annotation> annotation,
            boolean implicit) {
        return hasAnnotation(annotation.getCanonicalName(), implicit);
    }

    /**
     * Adds the given annotation, represented by an {@link AnnotationData},
     * to this class type at the location provided by the
     * {@link AnnotationData}.
     *
     * @param annotation the annotation to be added
     */
    public void annotate(AnnotationData annotation) {
        annotations.add(annotation);
        excludes.remove(annotation);
    }

    /**
     * Adds the given annotation, represented by an {@link AnnotationData},
     * to this class type at the location provided by the class
     *
     * @param annotation the annotation class to be added
     */
    public void annotate(Class<? extends Annotation> annotation) {
        annotate(annotation.getCanonicalName());
    }

    /**
     * Adds the given annotation, represented by an {@link AnnotationData},
     * to this class type at the location provided by the
     * annotation name
     *
     * @param annotation the annotation to add
     */
    public void annotate(CharSequence annotation) {
        AnnotationData annon = annotationFactory.createAnnotation(annotation, AnnotationLocation.RAW);
        annotations.add(annon);
    }

    /**
     * Forces the given annotation to be included among the annotations checked
     * by {@link AnnotatedTypeMirror#hasAnnotation}. This is useful for
     * treating class types as having an annotation even though they don't
     * (e.g., a primitive type that is implicitly {@code @NonNull}).
     *
     * @param annotation the annotation to include among this type's annotations
     */
    public void include(AnnotationData annotation) {
        includes.add(annotation);
        excludes.remove(annotation);
    }

    /**
     * Forces the given annotation to be included among the annotations 
     * checked
     * by {@link AnnotatedTypeMirror#hasAnnotation}. This is useful for
     * treating class types as having an annotation even though they don't
     * (e.g., a primitive type that is implicitly {@code @NonNull}).
     *
     * @param annotation the annotation class to include among this type's 
     * annotations
     */
    public void include(Class<? extends Annotation> annotation) {
        include(annotation.getCanonicalName());
    }

    /**
     * Forces the given annotation to be included among the annotations checked
     * by {@link AnnotatedTypeMirror#hasAnnotation}. This is useful for
     * treating class types as having an annotation even though they don't
     * (e.g., a primitive type that is implicitly {@code @NonNull}).
     *
     * @param annotation the annotation to include among this type's annotations
     */
    public void include(CharSequence annotation) {
        AnnotationData annon = annotationFactory.createAnnotation(annotation, AnnotationLocation.RAW);
        include(annon);
    }

    /**
     * Forces the given annotation to be skipped by {@code hasAnnotationAt}.
     * Useful in conjunction with annotation defaults (e.g., if variables are
     * {@code @NonNull} by default and the checker finds a {@code @Nullable}
     * variable, it can simply exclude {@code @NonNull} rather than handle
     * {@code @Nullable} as a special case).
     *
     * @param annotation the annotation to exclude from this type's annotations
     */
    public void exclude(AnnotationData annotation) {
        excludes.add(annotation);
        includes.remove(annotation);
    }

    /**
     * Forces the given annotation to be skipped by {@code hasAnnotationAt}.
     * Useful in conjunction with annotation defaults (e.g., if variables are
     * {@code @NonNull} by default and the checker finds a {@code @Nullable}
     * variable, it can simply exclude {@code @NonNull} rather than handle
     * {@code @Nullable} as a special case).
     *
     * @param annotation the annotation class to exclude from this type's annotations
     */
    public void exclude(Class<? extends Annotation> annotation) {
        exclude(annotation.getCanonicalName());
    }

    /**
     * Forces the given annotation to be skipped by {@code hasAnnotationAt}.
     * Useful in conjunction with annotation defaults (e.g., if variables are
     * {@code @NonNull} by default and the checker finds a {@code @Nullable}
     * variable, it can simply exclude {@code @NonNull} rather than handle
     * {@code @Nullable} as a special case).
     *
     * @param annotation the annotation name to exclude from this type's annotations
     */
    public void exclude(CharSequence annotation) {
        AnnotationData annon = annotationFactory.createAnnotation(annotation, AnnotationLocation.RAW);
        exclude(annon);
    }

    /* To ease transitions */
    public AnnotationLocation getPosition() {
        throw new RuntimeException("Not implemented yet!");
    }

    /**
     * Sets the Element associated with the value of this type, if one exists
     * 
     * @param elem the element of this type value
     */
    public void setElement(Element elem) {
        this.element = elem;
    }

    /**
     * @return  the {@code Element} of the value of this type
     */
    public Element getElement() {
        return element;
    }

    /**
     * Returns a shallow copy of this type.
     * 
     * @param annotation
     *            whether copy should have annotations
     */
    public AnnotatedTypeMirror getCopy(boolean annotation) {
        AnnotatedTypeMirror type = new AnnotatedTypeMirror(actualType, env);
        type.setElement(this.element);

        if (annotation) {
            for (AnnotationData annon : annotations)
                type.annotate(annon);
            for (AnnotationData annon : includes)
                type.include(annon);
            for (AnnotationData annon : excludes)
                type.exclude(annon);
        }

        return type;
    }

    /**
     * Represents a declared type (whether class or interface).
     */
    public static class AnnotatedDeclaredType extends AnnotatedTypeMirror {

        /** Parametrized Type Arguments **/
        final protected List<AnnotatedTypeMirror> typeArgs =
            new LinkedList<AnnotatedTypeMirror>();

        /** Supertype of this type **/
        protected AnnotatedTypeMirror superType;
        
        /** The interfaces that this type implements **/
        protected List<AnnotatedTypeMirror> interfaces;

        boolean isGeneric = false;
        
        /**
         * Constructor for this type
         * 
         * @param type  underlying kind of this type
         * @param env   the processing environment
         */
        public AnnotatedDeclaredType(TypeMirror type,
                ProcessingEnvironment env) {
            super(type, env);
            TypeElement elem =(TypeElement)env.getTypeUtils().asElement(type);
            isGeneric = !elem.getTypeParameters().isEmpty();
        }

        @Override
        public <R, P> R accept(AnnotatedTypeVisitor<R, P> v, P p) {
            return v.visitDeclared(this, p);
        }

        /**
         * Sets the type arguments on this type
         * @param t the type arguments
         */
        public void setTypeArguments(List<? extends AnnotatedTypeMirror> ts) {
            typeArgs.clear();
            typeArgs.addAll(ts);
        }

        /**
         * @return the type argument for this type
         */
        public List<? extends AnnotatedTypeMirror> getTypeArguments() {
            return Collections.unmodifiableList(typeArgs);
        }

        /**
         * Sets the supertype associated with this type
         * 
         * @param superType the super type of this type
         */
        public void setSupertype(AnnotatedTypeMirror superType) {
            this.superType = superType;
        }

        /**
         * @return the super type of this
         */
        public AnnotatedTypeMirror getSupertype() {
            return superType;
        }

        /**
         * Returns true if the type is generic, even if the type is erased
         * or used as a RAW type.
         * 
         * @returns true iff the type is generic
         */
        public boolean isGeneric() {
            return isGeneric;
        }
        /**
         * Sets the interfaces of this type
         * 
         * @param interfaces the interface types of this
         */
        public void setInterfaces(
                List<? extends AnnotatedTypeMirror> interfaces) {
            this.interfaces.clear();
            this.interfaces.addAll(interfaces);
        }
        
        /**
         * @return the interfaces of this type
         */
        public List<? extends AnnotatedTypeMirror> getInterfaces() {
            return Collections.unmodifiableList(interfaces);
        }

        @Override
        public AnnotatedDeclaredType getCopy(boolean annotation) {
            AnnotatedDeclaredType type =
                new AnnotatedDeclaredType(this.getUnderlyingType(), this.env);
            type.setElement(getElement());
            type.setEnclosingType(this.getEnclosingType());
            type.setInterfaces(getInterfaces());
            type.setSupertype(getSupertype());
            type.setTypeArguments(getTypeArguments());

            if (annotation) {
                for (AnnotationData annon : annotations)
                    type.annotate(annon);
                for (AnnotationData annon : includes)
                    type.include(annon);
                for (AnnotationData annon : excludes)
                    type.exclude(annon);
            }

            return type;
        }
    }

    /**
     * Represents a type of an executable. An executable is a method, constructor, or initializer.
     */
    public static class AnnotatedExecutableType extends AnnotatedTypeMirror {

        protected AnnotatedExecutableType(TypeMirror type,
                ProcessingEnvironment env) {
            super(type, env);
        }

        final private List<AnnotatedTypeMirror> paramTypes =
            new LinkedList<AnnotatedTypeMirror>();
        private AnnotatedTypeMirror receiverType;
        private AnnotatedTypeMirror returnType;
        final private List<AnnotatedTypeMirror> throwsTypes =
            new LinkedList<AnnotatedTypeMirror>();
        final private List<AnnotatedTypeVariable> typeVarTypes =
            new LinkedList<AnnotatedTypeVariable>();

        @Override
        public <R, P> R accept(AnnotatedTypeVisitor<R, P> v, P p) {
            return v.visitExecutable(this, p);
        }

        /**
         * Sets the parameter types of this executable type
         * @param params the parameter types
         */
        public void setParameterTypes(
                List<? extends AnnotatedTypeMirror> params) {
            paramTypes.clear();
            paramTypes.addAll(params);
        }

        /**
         * @return the parameter types of this executable type
         */
        public List<? extends AnnotatedTypeMirror> getParameterTypes() {
            return Collections.unmodifiableList(paramTypes);
        }

        /**
         * Sets the return type of this executable type
         * @param returnType    the return type
         */
        public void setReturnType(AnnotatedTypeMirror returnType) {
            this.returnType = returnType;
        }

        /**
         * @return the return type of this executable type
         */
        public AnnotatedTypeMirror getReturnType() {
            return returnType;
        }

        /**
         * Sets the receiver type on this executable type
         * @param receiverType the receiver type
         */
        public void setReceiverType(AnnotatedTypeMirror receiverType) {
            this.receiverType = receiverType;
        }

        /**
         * @return the receiver type of this executable type
         */
        public AnnotatedTypeMirror getReceiverType() {
            return receiverType;
        }

        /**
         * Sets the thrown types of this executable type
         * 
         * @param thrownTypes the thrown types
         */
        public void setThrownTypes(
                List<? extends AnnotatedTypeMirror> thrownTypes) {
            this.throwsTypes.clear();
            this.throwsTypes.addAll(thrownTypes);
        }

        /**
         * @return the thrown types of this executable type
         */
        public List<? extends AnnotatedTypeMirror> getThrownTypes() {
            return Collections.unmodifiableList(throwsTypes);
        }

        /**
         * Sets the type variables associated with this executable type
         * 
         * @param types the type variables of this executable type
         */
        public void setTypeVariables(List<AnnotatedTypeVariable> types) {
            typeVarTypes.clear();
            typeVarTypes.addAll(types);
        }

        /**
         * @return the type variables of this executable type, if any
         */
        public List<AnnotatedTypeVariable> getTypeVariables() {
            return Collections.unmodifiableList(typeVarTypes);
        }

        @Override
        public AnnotatedExecutableType getCopy(boolean annotation) {
            AnnotatedExecutableType type =
                new AnnotatedExecutableType(getUnderlyingType(), env);
            type.setElement(getElement());
            type.setParameterTypes(getParameterTypes());
            type.setReceiverType(getReceiverType());
            type.setReturnType(getReturnType());
            type.setThrownTypes(getThrownTypes());
            type.setTypeVariables(getTypeVariables());

            if (annotation) {
                for (AnnotationData annon : annotations)
                    type.annotate(annon);
                for (AnnotationData annon : includes)
                    type.include(annon);
                for (AnnotationData annon : excludes)
                    type.exclude(annon);
            }

            return type;
        }
    }

    /**
     * Represents Array types in java. A multidimensional array type is 
     * represented as an array type whose component type is also an 
     * array type.
     */
    public static class AnnotatedArrayType extends AnnotatedTypeMirror {

        protected AnnotatedArrayType(TypeMirror type,
                ProcessingEnvironment env) {
            super(type, env);
        }

        /** The component type of this array type */
        private AnnotatedTypeMirror componentType;

        @Override
        public <R, P> R accept(AnnotatedTypeVisitor<R, P> v, P p) {
            return v.visitArray(this, p);
        }

        /**
         * Sets the component type of this array
         * 
         * @param type the component type
         */
        public void setComponentType(AnnotatedTypeMirror type) {
            this.componentType = type;
        }

        /**
         * @return the component type of this array
         */
        public AnnotatedTypeMirror getComponentType() {
            return componentType;
        }


        @Override
        public AnnotatedArrayType getCopy(boolean annotation) {
            AnnotatedArrayType type = new AnnotatedArrayType(actualType, env);

            type.setElement(getElement());
            type.setComponentType(getComponentType());

            if (annotation) {
                for (AnnotationData annon : annotations)
                    type.annotate(annon);
                for (AnnotationData annon : includes)
                    type.include(annon);
                for (AnnotationData annon : excludes)
                    type.exclude(annon);
            }

            return type;
        }

    }

    /**
     * Represents a type variable. A type variable may be explicitly declared by
     * a type parameter of a type, method, or constructor. A type variable may 
     * also be declared implicitly, as by the capture conversion of a wildcard 
     * type argument (see chapter 5 of The Java Language Specification, Third 
     * Edition).
     *
     */
    public static class AnnotatedTypeVariable extends AnnotatedTypeMirror {

        protected AnnotatedTypeVariable(TypeMirror type,
                ProcessingEnvironment env) {
            super(type, env);
        }

        /** The element of the type variable **/
        private Element typeVarElement;
        /** The lower bound of the type variable **/
        private AnnotatedTypeMirror lowerBound;
        /** The upper bound of the type variable **/
        private AnnotatedTypeMirror upperBound;

        public <R, P> R accept(AnnotatedTypeVisitor<R, P> v, P p) {
            return v.visitTypeVariable(this, p);
        }

        /**
         * Set the element of the type variable
         * 
         * @param element   the type variable element
         */
        public void setTypeVariableElement(Element element) {
            this.typeVarElement = element;
        }

        /**
         * @return the element of the type variable in this
         */
        public Element getTypeVariableElement() {
            return typeVarElement;
        }

        /**
         * Set the lower bound of this variable type
         * 
         * @param type  the lower bound type
         */
        public void setLowerBound(AnnotatedTypeMirror type) {
            this.lowerBound = type;
        }

        /**
         * @return the lower bound of this variable type
         */
        public AnnotatedTypeMirror getLowerBound() {
            return lowerBound;
        }

        /**
         * Set the upper bound of this variable type
         * @param type the upper bound type
         */
        public void setUpperBound(AnnotatedTypeMirror type) {
            this.upperBound = type;
        }

        /**
         * @return the upper bound type of this type variable
         */
        public AnnotatedTypeMirror getUpperBound() {
            return upperBound;
        }

        @Override
        public AnnotatedTypeVariable getCopy(boolean annotation) {
            AnnotatedTypeVariable type =
                new AnnotatedTypeVariable(actualType, env);

            type.setElement(getElement());
            type.setLowerBound(getLowerBound());
            type.setTypeVariableElement(getTypeVariableElement());
            type.setUpperBound(getUpperBound());

            if (annotation) {
                for (AnnotationData annon : annotations)
                    type.annotate(annon);
                for (AnnotationData annon : includes)
                    type.include(annon);
                for (AnnotationData annon : excludes)
                    type.exclude(annon);
            }

            return type;
        }
    }
}
