package checkers.util;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

import javax.annotation.processing.ProcessingEnvironment;
import javax.lang.model.element.*;
import javax.lang.model.type.TypeKind;

import checkers.types.AnnotatedTypeMirror;
import checkers.types.AnnotationData;
import checkers.types.AnnotatedTypeVisitor;
import checkers.types.SimpleAnnotatedTypeVisitor;
import checkers.types.AnnotatedTypeMirror.*;

/**
 * Utility methods for operating on {@code AnnotatedTypeMirror}. This
 * class mimics the class {@see javax.lang.model.util.Types}.
 */
public class AnnotatedTypes {

    private ProcessingEnvironment env;

    /**
     * Constructor for {@code AnnotatedTypeUtils}
     * 
     * @param env
     *            the processing environment for this round
     */
    public AnnotatedTypes(ProcessingEnvironment env) {
        this.env = env;
    }

    /**
     * Returns the most specific base type of {@code t} that start
     * with the given {@code Element}
     * 
     * @param t         a type
     * @param element   an element
     * @return the base type of t of the given element
     */
    public AnnotatedTypeMirror asSuper(AnnotatedTypeMirror type,
            Element element) {
        switch (type.getKind()) {
        case DECLARED:
            AnnotatedDeclaredType t = (AnnotatedDeclaredType) type;
            // If visited Element is the desired one, we are done
            if (t.getElement().equals(element))
                return t;

            // Visit the superclass first!
            AnnotatedTypeMirror st = t.getSupertype();
            if (st.getKind() == TypeKind.DECLARED) {
                AnnotatedTypeMirror x = asSuper(st, element);
                if (x != null)
                    return x;
            }

            // Check if defined by interfaces
            if (element.getKind().isInterface()) {
                for (AnnotatedTypeMirror si : t.getInterfaces()) {
                    AnnotatedTypeMirror x = asSuper(si, element);
                    if (x != null)
                        return x;
                }
            }

            // Couldn't find the desired base type
            return null;
        case ARRAY:
            // Check if array component is subtype of the element
            // first
            return isSubtype(type, getAnnotatedType(element))
                    ? getAnnotatedType(element) : null;
        case TYPEVAR:
            // Operate on the upper bound
            return asSuper(((AnnotatedTypeVariable) type).getUpperBound(),
                    element);
        default:
            // No need to do super
            return type;
        }
    }

    /**
     * Return the base type of t or any of its outer types that starts
     * with the given Element. If none exists, return null.
     * 
     * @param t     a type
     * @param elem   a symbol
     */
    public AnnotatedTypeMirror asOuterSuper(AnnotatedTypeMirror t,
            Element elem) {
        switch (t.getKind()) {
        case DECLARED:
            do {
                // Search among supers for a desired supertype
                AnnotatedTypeMirror s = asSuper(t, elem);
                if (s != null)
                    return s;
                // if not found immediately, try enclosing type
                // like A in A.B
                t = t.getEnclosingType();
            } while (t.getKind() == TypeKind.DECLARED);
            return null;
        case ARRAY:
            return isSubtype(t, getAnnotatedType(elem))
                    ? getAnnotatedType(elem) : null;
        case TYPEVAR:
            return asSuper(t, elem);
        default:
            return null;
        }
    }

    /**
     * This method checks for subtyping according to Java rules not
     * with respect annotated types.
     * 
     * @param t1    a type
     * @param t2    a type
     * @return true iff t1 is a subtype of t2
     */
    private boolean isSubtype(AnnotatedTypeMirror t1, AnnotatedTypeMirror t2) {
        // FOR NOW: We just want to test for regular testing
        return env.getTypeUtils().isSubtype(t1.getUnderlyingType(),
                t2.getUnderlyingType());
    }

    /**
     * A Place-holder for a method that returns the appropriate
     * {@code AnnotatedType} for the given element
     * 
     * @param elem
     *            an element
     * @return the type associated with the element
     */
    public AnnotatedTypeMirror getAnnotatedType(Element elem) {
        throw new RuntimeException("Not Implemented yet!");
    }

    /**
     * The type of given symbol, seen as a member of t.
     * 
     * @param t    a type
     * @param sym  a symbol
     */
    public AnnotatedTypeMirror asMemberOf(AnnotatedTypeMirror t, Element elem) {
        if (ElementUtils.isStatic(elem))
            return getAnnotatedType(elem);

        // For Type Variable, operate on the upper
        if (t.getKind() == TypeKind.TYPEVAR)
            return asMemberOf(((AnnotatedTypeVariable) t).getUpperBound(),
                    elem);

        // I cannot think of why it wouldn't be a declared type!
        // Defensive Programming
        if (t.getKind() != TypeKind.DECLARED) {
            return getAnnotatedType(elem);
        }

        //
        // Basic Algorithm:
        // 1. Find the owner of the element
        // 2. Find the base type of owner (e.g. type of owner as supertype 
        //      of passed type)
        // 3. Subsitute for type variables if any exist
        TypeElement owner = ElementUtils.enclosingClass(elem);
        AnnotatedDeclaredType ownerType =
            (AnnotatedDeclaredType) getAnnotatedType(owner);

        // TODO: Potential bug if Raw type is used
        if (ElementUtils.isStatic(elem) || owner.getTypeParameters().isEmpty())
            return getAnnotatedType(elem);

        AnnotatedDeclaredType base =
            (AnnotatedDeclaredType) asOuterSuper(t, owner);

        if (base == null)
            return getAnnotatedType(elem);

        List<? extends AnnotatedTypeMirror> ownerParams = 
            ownerType.getTypeArguments();
        List<? extends AnnotatedTypeMirror> baseParams =
            base.getTypeArguments();
        if (!ownerParams.isEmpty()) {
            if (baseParams.isEmpty())
                return erasure(getAnnotatedType(elem));
            else
                return subst(getAnnotatedType(elem), ownerParams, baseParams);
        }

        return getAnnotatedType(elem);

    }

    /**
     * Returns the erasure type of the passed type, according to JLS
     * specifications.
     * 
     * @param type the type to be erasured
     * @return  the erasure of the passed type
     */
    public AnnotatedTypeMirror erasure(AnnotatedTypeMirror type) {
        switch (type.getKind()) {
        case TYPEVAR:
            // |T extends A&B| = |A|
        case WILDCARD:
            // |? extends A&B| = |A|
            return erasure(((AnnotatedTypeVariable)type).getUpperBound());
        case ARRAY:
            // | T[ ] | = |T| [ ]
            AnnotatedArrayType at = ((AnnotatedArrayType)type).getCopy(true);
            AnnotatedTypeMirror ct = erasure(at.getComponentType());
            at.setComponentType(ct);
            return at;
        case DECLARED:
            // 1. |G<T_1, ..., T_n>| = |G|
            // 2. |T.C| = |T|.C
            AnnotatedDeclaredType dt = (AnnotatedDeclaredType)type;
            if (!dt.getTypeArguments().isEmpty()) {
                // Handle case 1.
                AnnotatedDeclaredType rType = dt.getCopy(true);
                rType.setTypeArguments(Collections.<AnnotatedTypeMirror> emptyList());
                return erasure(rType);
            } else if (dt.getEnclosingType().getKind() != TypeKind.NONE) {
                // Handle case 2
                // TODO: Test this
                AnnotatedDeclaredType rType = dt.getCopy(true);
                AnnotatedTypeMirror et = dt.getEnclosingType();
                rType.setEnclosingType(erasure(et));
                return rType;
            } else {
                return dt;
            }
        default:
            return type;
        }
    }

    /**
     * Returns a new type, a copy of the passed {@code t}, with all
     * instances of {@code from} type substituted with their correspondents
     * in {@code to} and return the substituted in type.
     * 
     * @param t     the type
     * @param from  the from types
     * @param to    the to types
     * @return  the new type after substitutions
     */
    public AnnotatedTypeMirror subst(AnnotatedTypeMirror t,
            List<? extends AnnotatedTypeMirror> from,
            List<? extends AnnotatedTypeMirror> to) {
        assert from.size() == to.size();
        Map<AnnotatedTypeMirror, AnnotatedTypeMirror> mappings =
            new HashMap<AnnotatedTypeMirror, AnnotatedTypeMirror>();

        for (int i = 0; i < from.size(); ++i) {
            mappings.put(from.get(i), to.get(i));
        }
        return subst(t, mappings);
    }

    /**
     * Returns a new type, a copy of the passed {@code t}, with all
     * instances of {@code mappings.keySet()} type substituted with 
     * their correspondents in {@code to} and return the substituted 
     * in type.
     * 
     * @param t     the type
     * @param mappings  the mapping for the types
     * @return  the new type after substitutions
     */
    public AnnotatedTypeMirror subst(
            AnnotatedTypeMirror t,
            Map<? extends AnnotatedTypeMirror, ? extends AnnotatedTypeMirror> mappings) {
        return new Subst(mappings).subst(t);
    }

    /**
     * Returns a new type, a copy of the passed {@code t}, with all
     * instances of {@code mappings.keySet()} type substituted with 
     * their correspondents in {@code to} and return the substituted 
     * in type.
     * 
     * @param t     the type
     * @param mappings  the mapping for the types
     * @return  the new type after substitutions
     */
    public List<? extends AnnotatedTypeMirror> subst(
            List<? extends AnnotatedTypeMirror> ts,
            Map<? extends AnnotatedTypeMirror, ? extends AnnotatedTypeMirror> mappings) {
        return new Subst(mappings).subst(ts);
    }

    /**
     * Helper visitor class for the substitutions
     */
    private static class Subst extends
            SimpleAnnotatedTypeVisitor<AnnotatedTypeMirror, Void> {

        Map<? extends AnnotatedTypeMirror, ? extends AnnotatedTypeMirror> mappings;

        public Subst(
                Map<? extends AnnotatedTypeMirror, ? extends AnnotatedTypeMirror> mappings) {
            this.mappings =
                new HashMap<AnnotatedTypeMirror, AnnotatedTypeMirror>(
                        mappings);
        }

        AnnotatedTypeMirror subst(AnnotatedTypeMirror type) {
            if (mappings.isEmpty())
                return type;
            else
                return visit(type, null);
        }

        List<? extends AnnotatedTypeMirror> subst(
                List<? extends AnnotatedTypeMirror> types) {
            if (mappings.isEmpty())
                return types;
            else {
                List<AnnotatedTypeMirror> results =
                    new ArrayList<AnnotatedTypeMirror>();
                for (AnnotatedTypeMirror t : types)
                    results.add(subst(t));
                return results;
            }
        }

        @Override
        public AnnotatedTypeMirror visitType(AnnotatedTypeMirror t, Void p) {
            return t;
        }

        @Override
        public AnnotatedExecutableType visitExecutable(
                AnnotatedExecutableType t, Void p) {
            List<? extends AnnotatedTypeMirror> args =
                subst(t.getParameterTypes());
            AnnotatedTypeMirror receiverType = subst(t.getReceiverType());
            AnnotatedTypeMirror returnType = subst(t.getReturnType());
            List<? extends AnnotatedTypeMirror> thrownTypes =
                subst(t.getThrownTypes());

            AnnotatedExecutableType copy = t.getCopy(true);
            copy.setParameterTypes(args);
            copy.setReceiverType(receiverType);
            copy.setReturnType(returnType);
            copy.setThrownTypes(thrownTypes);

            return copy;
        }

        @Override
        public AnnotatedArrayType visitArray(AnnotatedArrayType t, Void p) {
            AnnotatedArrayType array = t.getCopy(true);
            array.setComponentType(subst(t.getComponentType()));
            return array;
        }

        @Override
        public AnnotatedTypeMirror visitTypeVariable(AnnotatedTypeVariable t,
                Void p) {
            // FIXME: This has bugs!
            // .equals may not be the way to go
            if (!mappings.containsKey(t))
                return t;

            AnnotatedTypeMirror copy = mappings.get(t).getCopy(true);
            for (AnnotationData annon : t.getAnnotations(false))
                copy.annotate(annon);
            return copy;
        }
    }

    /**
     * Returns a deep copy of the passed type
     * 
     * @param type  the annotated type to be copied
     * @return a deep copy of the passed type
     */
    public AnnotatedTypeMirror deepCopy(AnnotatedTypeMirror type) {
        return deepCopier.visit(type, true);
    }

    /**
     * A Helper visitor for deep copying
     */
    private AnnotatedTypeVisitor<AnnotatedTypeMirror, Boolean> deepCopier =
        new AnnotatedTypeVisitor<AnnotatedTypeMirror, Boolean>() {

            @Override
            public AnnotatedTypeMirror visit(AnnotatedTypeMirror t) {
                return visit(t, null);
            }

            @Override
            public AnnotatedTypeMirror visit(AnnotatedTypeMirror type,
                    Boolean p) {
                return type.accept(this, p);
            }

            @Override
            public AnnotatedArrayType visitArray(AnnotatedArrayType type,
                    Boolean p) {
                AnnotatedArrayType copy = type.getCopy(p);
                copy.setComponentType(deepCopier.visit(copy.getComponentType(),
                        p));
                return copy;
            }

            @Override
            public AnnotatedTypeMirror visitDeclared(
                    AnnotatedDeclaredType type, Boolean p) {
                AnnotatedDeclaredType copy = type.getCopy(p);
                type.setEnclosingType(deepCopier.visit(type
                        .getEnclosingType(), p));
                return copy;
            }

            @Override
            public AnnotatedExecutableType visitExecutable(
                    AnnotatedExecutableType type, Boolean p) {
                AnnotatedExecutableType copy = type.getCopy(p);

                List<AnnotatedTypeMirror> params =
                    new LinkedList<AnnotatedTypeMirror>();
                for (AnnotatedTypeMirror annotatedParam : type
                        .getParameterTypes()) {
                    params.add(deepCopier.visit(annotatedParam, true));
                }
                copy.setParameterTypes(params);

                copy.setReceiverType(deepCopier.visit(type.getReceiverType(),
                        true));
                copy.setReturnType(deepCopier.visit(type.getReturnType(),
                        true));

                List<AnnotatedTypeMirror> throwsList =
                    new LinkedList<AnnotatedTypeMirror>();
                for (AnnotatedTypeMirror throwsType : type.getThrownTypes())
                    throwsList.add(deepCopier.visit(throwsType, true));

                return copy;
            }

            @Override
            public AnnotatedTypeMirror visitType(AnnotatedTypeMirror type,
                    Boolean p) {
                return type.getCopy(p);
            }

            @Override
            public AnnotatedTypeMirror visitTypeVariable(
                    AnnotatedTypeVariable type, Boolean p) {
                AnnotatedTypeVariable copy = type.getCopy(p);

                copy.setLowerBound(deepCopier.visit(copy.getLowerBound(), 
                        true));
                copy.setUpperBound(deepCopier.visit(copy.getUpperBound(),
                        true));
                return copy;
            }
        };
}
