package checkers.util;

import checkers.types.*;

import com.sun.source.tree.*;

import java.util.*;

import javax.annotation.processing.*;
import javax.lang.model.element.*;
import javax.lang.model.type.*;
import javax.lang.model.util.*;

/**
 * Utilities for working with annotated generic types, and particularly,
 * applying annotations from annotated type arguments to type variables.
 */
public class GenericsUtils {

    public static boolean DEBUG = false;

    private final ProcessingEnvironment env;
    private final Types types;
    private final Elements elements;
    private final AnnotatedTypeFactory factory;

    /**
     * Creates a {@link GenericsUtils} instance.
     *
     * @param env the {@link ProcessingEnvironment} to use for type utilities
     * @param factory the {@link AnnotatedTypeFactory} to use for getting
     *                annotated types
     */
    public GenericsUtils(ProcessingEnvironment env,
            AnnotatedTypeFactory factory) {
        this.env = env;
        this.types = env.getTypeUtils();
        this.elements = env.getElementUtils();
        this.factory = factory;
    }

    /**
     * A visitor for types that tracks locations as it descends into generic
     * types.
     */
    public static class LocationVisitor<R,P> 
        extends SimpleTypeVisitor6<R, P> {
        
        private List<Integer> loc = new ArrayList<Integer>();
        
        /**
         * @return the location of the currently visited type
         */
        public AnnotationLocation getCurrentLocation() {
            return AnnotationLocation.fromList(loc);
        }
        
        protected final R scan(TypeMirror t, P p) {
            return visit(t, p);
        }
        
        protected R reduce(R r1,R r2) {
            return (r1 != null) ? r1 : r2;
        }
        
        protected R scanAndReduce(TypeMirror t, P p, R r) {
            return reduce(scan(t, p), r);
        }
        
        protected R scan(Iterable<? extends TypeMirror> types, P p) {
            boolean first = true;
            R r = null;
            for (TypeMirror t : types) {
                r = first ? scan(t, p) : scanAndReduce(t, p, r);
            }
            return r;
        }
        
        /**
         * This method needs to be called to visit 
         * type parameters.
         */
        @Override
        public R visitDeclared(DeclaredType t, P p) {
            final int depth = loc.size();
            loc.add(0);
            
            int i = 0;
            R r = super.visitDeclared(t, p);
            for (TypeMirror arg : t.getTypeArguments()) {
                loc.set(depth, i++);
                r = scanAndReduce(arg, p, r);
            }
            try {
                return r;
            } finally {
                // calls remove(int) not remove(Object)
                loc.remove(depth);
            }
        }
    }
    
    /**
     * A visitor for types that maps types to locations, used for matching
     * annotation locations to types.
     */
    class MatchAnnotationsVisitor extends LocationVisitor<Void, Void> {

        private final Map<AnnotationLocation, TypeMirror> locs =
            new HashMap<AnnotationLocation, TypeMirror>();

        @Override
        public Void visitDeclared(DeclaredType t, Void p) {
            locs.put(getCurrentLocation(), t);
            return super.visitDeclared(t, p);
        }

        @Override
        public Void visitTypeVariable(TypeVariable t, Void p) {
            locs.put(getCurrentLocation(), t);
            return super.visitTypeVariable(t, p);
        }

        public Map<AnnotationLocation, TypeMirror> getTypeLocations() {
            return Collections.unmodifiableMap(locs);
        }

    }

    /**
     * A visitor for types that matches type variables to annotations.
     */
    class ApplyAnnotationsVisitor extends LocationVisitor<Void, Void> {

        public final Set<AnnotationData> annos = new HashSet<AnnotationData>();
        private final Map<TypeMirror, AnnotationData> matched;
        private final AnnotationFactory af;

        public ApplyAnnotationsVisitor(Map<TypeMirror, AnnotationData> matched,
                AnnotationFactory af) {
            this.matched = matched;
            this.af = af;
        }

        @Override
        public Void visitTypeVariable(TypeVariable t, Void p) {

            // Get the annotation data for this type variable.
            AnnotationData ad = matched.get(t);

            if (ad != null) {
                annos.add(
                        // Create an annotation with the correct name at *this*
                        // location (not the annotation's original location).
                        af.createAnnotation(
                            ((TypeElement)types.asElement(ad.getType())).getQualifiedName(),
                            getCurrentLocation()));
            }

            return super.visitTypeVariable(t, p);
        }

    }

    /**
     * Use the MatchAnnotationsVisitor to match annotations with a certain
     * location to types with the same location. Often {@code type} and {@code
     * source} will be the type and {@link AnnotatedClassType} for a common
     * {@link Element}, respectively.
     *
     * @param type the type to visit
     * @param source the source of annotations for that type
     * @return a mapping from types to annotations that have the same
     *         annotations
     */
    Map<TypeMirror, AnnotationData> matchAnnotations(TypeMirror type,
            AnnotatedClassType source) {

        Map<TypeMirror, AnnotationData> typevars =
            new HashMap<TypeMirror, AnnotationData>();

        Map<AnnotationLocation, AnnotationData> annotations =
            mapLocations(source.getAnnotationData(true));

        assert type instanceof DeclaredType;

        DeclaredType dt = (DeclaredType)type;

        // Get the this type mirror and its supertypes for checking.
        List<TypeMirror> tms = new LinkedList<TypeMirror>();
        tms.add(dt);
        tms.addAll(types.directSupertypes(dt));

        for (TypeMirror tm : tms) {

            // Get the type as used and the type as declared. atm will be
            // something like Set<String> and gtm will be something like
            // Set<E>.
            DeclaredType atm = (DeclaredType)tm;
            DeclaredType gtm = (DeclaredType)types.asElement(tm).asType();

            // Get types (including type vars) by their locations.
            Map<AnnotationLocation, TypeMirror> typeLocations;
            {
                MatchAnnotationsVisitor mav = new MatchAnnotationsVisitor();
                mav.visit(gtm, null);
                typeLocations = mav.getTypeLocations();
            }

            for (Map.Entry<AnnotationLocation, AnnotationData> entry :
                    annotations.entrySet()) {

                AnnotationLocation key = entry.getKey();
                AnnotationData value = entry.getValue();

                // Get the name of the annotation.
                CharSequence annoName;
                {
                    TypeMirror annoType = value.getType();
                    TypeElement annoElt = (TypeElement)types.asElement(annoType);
                    annoName = annoElt.getQualifiedName();
                }

                AnnotationFactory af = new AnnotationFactory(env);

                // If the annotation location also maps to a type mirror, put
                // it in the result map.
                if (typeLocations.containsKey(key)) {
                    AnnotationData mapped = af.createAnnotation(annoName,
                            AnnotationLocation.RAW);
                    typevars.put(typeLocations.get(key), mapped);
                }
            }

        }

        return typevars;
    }

    /**
     * Uses an ApplyAnnotationsVisitor to determine the annotations on the
     * given type. Annotations are applied from a given set of matches, which
     * can be determined by {@link GenericsUtils#matchAnnotations}.
     *
     * @param type the type to apply annotations to
     * @param matches the annotations to apply
     * @return the annotations on {@code type} as determined by {@code matches}
     */
    Set<AnnotationData> applyAnnotations(TypeMirror type,
            Map<TypeMirror, AnnotationData> matches) {
        AnnotationFactory af = new AnnotationFactory(this.env);
        ApplyAnnotationsVisitor aav = new ApplyAnnotationsVisitor(matches, af);
        aav.visit(type, null);
        return aav.annos;
    }

    /**
     * @param annotations a set of annotations
     * @return {@code annotations} in a map keyed by their locations
     */
    static Map<AnnotationLocation, AnnotationData>
        mapLocations(Set<AnnotationData> annotations) {

        Map<AnnotationLocation, AnnotationData> byLoc =
            new HashMap<AnnotationLocation, AnnotationData>();

        for (AnnotationData annotation : annotations)
            byLoc.put(annotation.getLocation(), annotation);
        return Collections.unmodifiableMap(byLoc);
    }

    public Set<AnnotationData> annotationsFor(TypeMirror type, Element elt) {
        return annotationsFor(type, elt, Collections.EMPTY_MAP);
    }


    /*
     * Determines the annotations for a type (which presumably is a type
     * variable or contains type variables) based on the given expression and
     * annotation source. The expression is that which is a use of the type
     * variables in type; the annotation source is usually the {@link
     * AnnotatedClassType} of the expression's element.
     */
    public Set<AnnotationData> annotationsFor(TypeMirror type, Element elt,
            Map<TypeMirror, AnnotationData> moreMappings) {

        if (DEBUG) {
            System.out.println("**    TYPE: " + type + " / " + type.getKind());
            System.out.println("** ELEMENT: " + elt + " / " + elt.getKind());
        }

        TypeMirror expression = elt.asType();
        AnnotatedClassType source = factory.getClass(elt);

        if (expression.getKind() == TypeKind.EXECUTABLE) {
            ExecutableType method = (ExecutableType)expression;
            expression = method.getReturnType();
        }

        if (expression.getKind() != TypeKind.DECLARED)
            return Collections.EMPTY_SET;

        Map<TypeMirror, AnnotationData> mappings =
            this.matchAnnotations(expression, source);
        mappings.putAll(moreMappings);
        
        Set<AnnotationData> result =
            this.applyAnnotations(type, mappings);

        if (DEBUG) {
            System.out.println("--     RESULT: " + result);
        }

        return result;
    }

    /**
     * Determines the iterated type in an enhanced for loop. For instance, if
     * iterating over a {@code et&lt;String&gt;}, the method returns the
     * declared type of {@code Set&lt;String&gt;.iterator().next()}, which is
     * the type of {@code Set&lt;E&gt;.iterator.next()}, which is {@code E}.
     *
     * @param tree the enhanced for loop
     * @return the type of the iterated argument
     */
    public TypeMirror iteratedType(EnhancedForLoopTree tree) {

        // This method makes the distinction between a two different kinds of
        // types. There are types "as declared", such as "Iterable<T>" or
        // "Set<E>"; there are types "as used", such as "Set<String>" or
        // "Iterable<String>", and also "Iterable<E>" in "Set<E> extends
        // Iterable<E>".
        //
        // Really, what we're doing here is taking a type as instantiated --
        // Set<String> -- that we're iterating over, finding its type as
        // declared -- Set<E> -- finding the Iterable type as used --
        // Iterable<E> (not Iterable<T>, as it's declared) -- and getting the
        // return value of "Iterable.iterator().next()", which in this case is
        // E.
        //
        // Once we have done this, we can use GenericsUtils to match type
        // variables to get the annotations on "Iterable.iterator().next().
        // This is why we want to get Iterable<E> and eventually E, not
        // Iterable<T> and eventually T -- GenericsUtils can find that
        // E=@NonNull String for Set<@NonNull String>.

        // Get the element for the for expression.
        Element exprElt;

        if (tree.getExpression().getKind() == Tree.Kind.NEW_CLASS)
            exprElt = InternalUtils.symbol(((NewClassTree)(tree.getExpression())).getIdentifier());
        else
            exprElt = InternalUtils.symbol(tree.getExpression());

        // Get the raw type for Iterable, to compare against later.
        TypeMirror iterableRawType;
        {
            TypeElement iterableTypeElt =
                elements.getTypeElement("java.lang.Iterable");
            iterableRawType = types.erasure(iterableTypeElt.asType());
        }

        // Get the supertypes of the type of the expression as they are used.
        Set<TypeMirror> superTypes;
        {
           TypesUtils types2 = new TypesUtils(this.env);

           if (exprElt instanceof ExecutableElement) {
               superTypes = types2.superTypesAsTypeMirrors(
                       (TypeElement)types.asElement(((ExecutableElement)exprElt).getReturnType()));
           } else {
               // Get type of the for expression as it has been declared (Set<E>
               // instead of Set<@NonNull String>.
               Element exprClass = types.asElement(exprElt.asType());
               if (exprClass == null)
                   return null;
               assert exprClass instanceof TypeElement : exprClass.getKind();

               // Get the supertypes as they are used (Iterable<E> instead of
               // Iterable<T>).
               superTypes = types2.superTypesAsTypeMirrors((TypeElement)exprClass);
           }
        }

        // Find the supertype that is the same as Iterable<T> when erased --
        // the result will be something like Iterable<E> or Iterable<A>.
        TypeMirror iterableType = null;
        for (TypeMirror superType : superTypes) {
            if (types.isSameType(types.erasure(superType), iterableRawType)) {
                iterableType = superType;
                break;
            }
        }

        assert iterableType != null;

        // Get the Iterable type as used, as an element.
        Element iterableTypeAsElt = types.asElement(iterableType);

        // Get the element for the Iterable.iterator() method.
        ExecutableElement iteratorMethod =
            ElementFilter.methodsIn(iterableTypeAsElt.getEnclosedElements()).get(0);

        // Get the type of the Iterable.iterator() method as used.
        ExecutableType iteratorMethodType =
            (ExecutableType)types.asMemberOf((DeclaredType)iterableType,
                                              iteratorMethod);

        // Get the element for the Iterator class as returned by the (as used)
        // Iterable type's iterator() method.
        TypeElement iteratorClass =
            (TypeElement)types.asElement(iteratorMethodType.getReturnType());

        // Get the type of the Iterator's .next() method.
        ExecutableType iteratorDotNext = null;
        for (ExecutableElement anIteratorMethod :
                ElementFilter.methodsIn(iteratorClass.getEnclosedElements())) {
            if (elements.getName("next").equals(anIteratorMethod.getSimpleName())) {
                iteratorDotNext = (ExecutableType)
                    types.asMemberOf((DeclaredType)iteratorMethodType.getReturnType(),
                            anIteratorMethod);
                break;
            }
        }

        assert iteratorDotNext != null;

        // Return the return type of Iterator.next().
        return iteratorDotNext.getReturnType();
    }
}
