package checkers.util;

import static javax.lang.model.util.ElementFilter.methodsIn;

import java.util.ArrayDeque;
import java.util.Collections;
import java.util.Deque;
import java.util.HashSet;
import java.util.Set;

import javax.annotation.processing.ProcessingEnvironment;
import javax.lang.model.element.*;
import javax.lang.model.type.TypeMirror;

/**
 * A Utilities class that helps with {@code Element}'s
 * 
 */
// TODO: This class needs significant restructuring
public final class TypesUtils {

    private ProcessingEnvironment env;

    /**
     * Constructor for TypeUtils.
     * 
     * @param environment
     *            the {@code ProcessingEnvironment} of the Annotation
     *            Type Processor
     */
    public TypesUtils(ProcessingEnvironment environment) {
        this.env = environment;
    }

    /**
     * A utility method that takes the element for a class/interface
     * type and returns a set of {@link TypeElement}s representing all
     * of the supertypes of that type.
     * 
     * @param subtype
     *            the element of the type for which all supertypes
     *            will be obtained
     * @return an unmodifiable set of supertypes for {@code subtype},
     */
    public Set<TypeElement> superTypes(TypeElement subtype) {

        Set<TypeElement> supertypes = new HashSet<TypeElement>();

        // Set up a stack containing the type mirror of subtype, which
        // is our starting point.
        Deque<TypeMirror> stack = new ArrayDeque<TypeMirror>();
        stack.push(subtype.asType());

        while (!stack.isEmpty()) {
            TypeMirror current = stack.pop();

            // For each direct supertype of the current type, if it
            // hasn't already been visited, push it onto the stack and
            // add it to our supertypes set.
            for (TypeMirror supertype : env.getTypeUtils()
                    .directSupertypes(current)) {
                if (!supertypes.contains(supertype)) {
                    stack.push(supertype);
                    supertypes.add((TypeElement) env.getTypeUtils()
                            .asElement(supertype));
                }
            }
        }

        return Collections.unmodifiableSet(supertypes);
    }

    /**
     * A utility method that takes the element for a class/interface
     * type and returns a set of {@link TypeMirror}s representing all
     * of the supertypes of that type.
     * 
     * @param subtype
     *            the element of the type for which all supertypes
     *            will be obtained
     * @return an unmodifiable set of supertypes for {@code subtype},
     */
    public Set<TypeMirror> superTypesAsTypeMirrors(TypeElement subtype) {

        Set<TypeMirror> supertypes = new HashSet<TypeMirror>();

        // Set up a stack containing the type mirror of subtype, which
        // is our starting point.
        Deque<TypeMirror> stack = new ArrayDeque<TypeMirror>();
        stack.push(subtype.asType());

        while (!stack.isEmpty()) {
            TypeMirror current = stack.pop();

            // For each direct supertype of the current type, if it
            // hasn't already been visited, push it onto the stack and
            // add it to our supertypes set.
            for (TypeMirror supertype : env.getTypeUtils()
                    .directSupertypes(current)) {
                if (!supertypes.contains(supertype)) {
                    stack.push(supertype);
                    supertypes.add(supertype);
                }
            }
        }

        return Collections.unmodifiableSet(supertypes);
    }

    /**
     * A utility method that takes a Method element and returns a set
     * of all elements that this method overrides (as
     * {@link ExecutableElement}s)
     * 
     * @param method
     *            the overriding method
     * @return an unmodifiable set of {@link ExecutableElements}
     *         representing the elements that method overrides
     */
    public Set<ExecutableElement> overriddenMethods(
            ExecutableElement method) {
        Set<TypeElement> supertypes =
            superTypes((TypeElement) method.getEnclosingElement());
        return overriddenMethods(method, supertypes);
    }

    /**
     * A utility method that takes the element for a method and the
     * set of all supertypes of the method's containing class and
     * returns the set of all elements that method overrides (as
     * {@link ExecutableElement}s).
     * 
     * @param method
     *            the overriding method
     * @param supertypes
     *            the set of supertypes to check for methods that are
     *            overriden by {@code method}
     * @return an unmodified set of {@link ExecutableElements}
     *         representing the elements that {@code} method overrides
     *         among {@code supertypes}
     */
    public Set<ExecutableElement> overriddenMethods(
            ExecutableElement method, Set<TypeElement> supertypes) {

        Set<ExecutableElement> overrides =
            new HashSet<ExecutableElement>();

        for (TypeElement supertype : supertypes) {

            // For all method in the supertype, add it to the set if
            // it overrides the given method.
            for (ExecutableElement supermethod : methodsIn(supertype
                    .getEnclosedElements())) {
                if (env.getElementUtils().overrides(method, supermethod,
                        supertype)) {
                    overrides.add(supermethod);
                    break;
                }
            }
        }

        return Collections.unmodifiableSet(overrides);
    }

    /**
     * Gets the innermost class enclosing the provided element.
     * 
     * @param element
     *            an element enclosed by a class, or a
     *            {@code TypeElement}
     * @return The innermost class enclosing the element
     */
    public TypeElement getClassOf(Element element) {
        Element elem = element;
        while (!((elem == null) || (elem.getKind().isClass()) || (elem
                .getKind().isInterface()))) {
            elem = elem.getEnclosingElement();
        }

        if (elem == null)
            return null;
        else {
            assert (elem instanceof TypeElement);
            return (TypeElement) elem;
        }
    }

    /**
     * Returns the qualified name of the inner most class enclosing
     * the provided {@code Element}
     * 
     * @param element
     *            an element enclosed by a class, or a
     *            {@code TypeElement}
     * @return The qualified {@code Name} of the innermost class
     *         enclosing the element
     */
    public Name getQualifiedClassName(Element element) {
        if (element == null) {
            return null;
        } else if (element.getKind() == ElementKind.PACKAGE) {
            PackageElement elem = (PackageElement) element;
            return elem.getQualifiedName();
        } else {
            TypeElement elem = getClassOf(element);
            if (elem == null)
                return null;
            else
                return elem.getQualifiedName();
        }
    }
}
